/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

export module Examples;

import Point;
import DefaultFraction;
import Fraction;
import LimitOrder;

import std;
using std::cout, std::format;

template <typename T>
concept Coutable = requires(std::ostream& os, const T& obj) 
{
	os << obj;
};

template <Coutable T>
void print_this(const T& t)
{
	cout << t << " ";
}

void vs()
{
	cout << "\n\n";
}

export void auto_deduced_examples()
{
	cout << "auto_deduced_examples():\n";

	int m = 0, n = 1;
	double x = 1.06, y = 8.74;
	std::string bass = "Rickenbacker", BASS = "RICKENBACKER";

	auto int_compare = m <=> n;
	auto real_compare = x <=> y;
	auto str_compare = bass <=> BASS;

	cout << format("int compare type: {}\nfloating compare type: {}\nstring compare type: {}\n",
		typeid(int_compare).name(), typeid(real_compare).name(), typeid(str_compare).name());
	vs();

	if (str_compare == std::strong_ordering::less)
		cout << "bass < BASS\n";
	else if (str_compare == std::strong_ordering::greater)
		cout << "bass > BASS\n";
	else
		cout << "bass == BASS\n";

	vs();
}

export void point_examples()
{
	cout << "Point class examples:\n";

	Point pt_01{1, 0, 6}, pt_02{8, 7, 4}, pt_03{4, 1, 9}, 
		  pt_04{4, 1, 9}, pt_05{4, 2, 5};

	print_this(pt_01);
	print_this(pt_02);
	print_this(pt_03);
	vs();

	cout << std::boolalpha;
	cout << pt_01 << "== " << pt_02 << "? " << (pt_01 == pt_02) << "\n";
	cout << pt_02 << "!= " << pt_03 << "? " << (pt_02 != pt_03) << "\n";
	cout << pt_03 << "== " << pt_04 << "? " << (pt_03 == pt_04) << "\n";
	vs();

	cout << pt_01 << "<= " << pt_02 << "? " << (pt_01 <= pt_02) << "\n";
	cout << pt_02 << "> " << pt_03 << "? " << (pt_02 > pt_03) << "\n";
	cout << pt_04 << "< " << pt_05 << "? " << (pt_04 < pt_05) << "\n";
	vs();
}

export void frac_default_spaceship_examples()
{
	using Fraction = DefaultFraction::Fraction;

	cout << "Fraction class with default == and <=>:\n";
	Fraction frac_01{1, 2}, frac_02{1, 5};

	cout << std::boolalpha;
	cout << frac_01 << "< " << frac_02 << "? " << (frac_01 < frac_02) << "\n";
	vs();
}

export void frac_implemented_spaceship_examples()
{
	cout << "Fraction class with user-defined <=>:\n";

	Fraction frac_01{1, 2}, frac_02{1, 5};
	cout << frac_01 << "< " << frac_02 << "? " << (frac_01 < frac_02) << "\n";
	vs();



	auto compare_type = [](const Fraction& frac_lhs, const Fraction& frac_rhs, 
		const std::strong_ordering& comp)
		{
			cout << "Comparison result: " << frac_lhs << " <=> " << frac_rhs << "\n";
			if (comp == std::strong_ordering::less)
			{
				cout << "strong_ordering::less" << "\n";
			}
			else if (comp == std::strong_ordering::greater)
			{
				cout << "strong_ordering::greater" << "\n";
			}
			else
			{
				cout << "strong_ordering::equivalent" << "\n";
			}

		};

	auto comp = frac_01 <=> frac_02;
	cout << "Type of comp: " << typeid(comp).name() << "\n";
	compare_type(frac_01, frac_02, comp);
	vs();

	comp = frac_02 <=> frac_01;
	compare_type(frac_02, frac_01, comp);
	vs();

	Fraction frac_03{1, 2};
	comp = frac_01 <=> frac_03;
	compare_type(frac_01, frac_03, comp);
	vs();
}

export void limit_order_examples()
{
	cout << "limit_order_examples()\n\n";

	auto compare = [](const std::partial_ordering& po)
	{
		if (po == std::partial_ordering::less) return " < ";
		else if (po == std::partial_ordering::equivalent) return " == ";
		else return " > ";
	};

	auto disp_result = [compare](int trade_id_lhs, int trade_id_rhs, const std::partial_ordering& po)
	{
		cout << format("Order {} {} Order {}", trade_id_lhs, compare(po), trade_id_rhs);
	};

	auto order_data = [](const LimitOrder& ord)
	{
		cout << format("Limit Order Data:\nBuy or Sell: {}\nBid/Offer: {}\nNumber of Shares: {}\nTimestamp: {}",
			ord.buy_or_sell(), ord.price(), ord.num_shares(), ord.time_stamp_string());
	};

	
	// Condition 1: Limit Price
	// Buy order
	LimitOrder buy_order_01{Buy_or_Sell::Buy, 100.0, 100, std::chrono::system_clock::now()};
	auto trade_id_01 = buy_order_01.trade_id();

	LimitOrder buy_order_02{Buy_or_Sell::Buy, 100.52, 100, std::chrono::system_clock::now()};
	auto trade_id_02 = buy_order_02.trade_id();

	order_data(buy_order_01);
	vs();
	order_data(buy_order_02);
	vs();

	auto comp = buy_order_01 <=> buy_order_02;
	disp_result(trade_id_01, trade_id_02, comp);
	vs();

	// Put another limit price comparison example with sell orders here
	LimitOrder sell_order_01{Buy_or_Sell::Sell, 95.0, 100, std::chrono::system_clock::now()};
	trade_id_01 = sell_order_01.trade_id();

	LimitOrder sell_order_02{Buy_or_Sell::Sell, 100.50, 100, std::chrono::system_clock::now()};
	trade_id_02 = sell_order_02.trade_id();

	order_data(sell_order_01);
	vs();
	order_data(sell_order_02);
	vs();

	disp_result(trade_id_01, trade_id_02, comp);
	vs();
	
	// Need round lot comparison examples
	LimitOrder buy_order_01_rl{Buy_or_Sell::Buy, 100.0, 75, std::chrono::system_clock::now()};
	trade_id_01 = buy_order_01_rl.trade_id();

	LimitOrder buy_order_02_rl{Buy_or_Sell::Buy, 100.52, 100, std::chrono::system_clock::now()};
	trade_id_02 = buy_order_02_rl.trade_id();

	order_data(buy_order_01_rl);
	vs();
	order_data(buy_order_02_rl);
	vs();

	comp = buy_order_01_rl <=> buy_order_02_rl;
	disp_result(trade_id_01, trade_id_02, comp);
	vs();

	// Timestamp comparison example:
	LimitOrder sell_order_01_ts{Buy_or_Sell::Sell, 100.0, 100, std::chrono::system_clock::now()};
	trade_id_01 = sell_order_01_ts.trade_id();

	// Do this to advance the timestamp by at least two seconds:
	std::this_thread::sleep_for(std::chrono::seconds(2));

	LimitOrder sell_order_02_ts{Buy_or_Sell::Sell, 100.0, 150, std::chrono::system_clock::now()};
	trade_id_02 = sell_order_02_ts.trade_id();

	order_data(sell_order_01_ts);
	vs();
	order_data(sell_order_02_ts);
	vs();

	comp = sell_order_01_ts <=> sell_order_02_ts;
	disp_result(trade_id_01, trade_id_02, comp);
	vs();
}