/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

module;
#include <cassert>

export module LimitOrder;

import std;

export enum class Buy_or_Sell
{
	Buy,
	Sell
};

export class LimitOrder
{
public:
	LimitOrder(Buy_or_Sell buy_or_sell, double price, int num_shares,
		std::chrono::time_point<std::chrono::system_clock> time_stamp) :
		buy_or_sell_{buy_or_sell}, price_{price}, num_shares_{num_shares}, 
		time_stamp_{time_stamp}
	{
		assert(num_shares > 0);
		assert(price > 0.0);
		++trade_id_;
	}

	bool operator ==(const LimitOrder& rhs) const
	{
		if(
			this->buy_or_sell_ == rhs.buy_or_sell_ &&
				std::abs(this->price_ - rhs.price_)
				< std::sqrt(std::numeric_limits<double>::epsilon()) &&
				this->num_shares_ == rhs.num_shares_ &&
				this->time_stamp_ == rhs.time_stamp_)
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	std::partial_ordering operator <=>(const LimitOrder& rhs) const
	{
		assert(this->buy_or_sell_ == rhs.buy_or_sell_);
	
		// Compare prices first, but flip depending on Buy or Sell
		std::partial_ordering cmp_price = (buy_or_sell_ == Buy_or_Sell::Buy)
			? (this->price_ <=> rhs.price_) // Case: Bid order  (higher price wins)
			: (rhs.price_ <=> price_);      // Case: Sell order (lower price wins)

		// First: Check if the absolute difference between the prices is 
		// within epsilon; if not, return the result of the <=> 
		// operator for the two order prices
		if (std::abs(this->price_ - rhs.price_) > epsilon_)
		{
			return cmp_price;
		}

		// Second: Prefer round lots (multiples of 100 shares)
		bool this_round_lot = (this->num_shares_ % 100) == 0;
		bool rhs_round_lot = (rhs.num_shares_ % 100) == 0;

		if (this_round_lot != rhs_round_lot)
		{		
			return this_round_lot ? std::partial_ordering::greater : std::partial_ordering::less;
		}

		// Third (and final comparison criterion): Earlier timestamp takes priority,
		// and hence "greater", i.e., higher in priority
		if (this->time_stamp_ < rhs.time_stamp_) return std::partial_ordering::greater;
		else if (this->time_stamp_ > rhs.time_stamp_) return std::partial_ordering::less;
		else return std::partial_ordering::equivalent;		// Also the overall default.
	}

	std::string buy_or_sell() const
	{
		return (buy_or_sell_ == Buy_or_Sell::Buy)? "Buy" : "Sell";
	}

	double price() const
	{
		return price_;
	}

	int num_shares() const
	{
		return num_shares_;
	}

	std::chrono::time_point<std::chrono::system_clock> time_stamp() const
	{
		return time_stamp_;
	}

	// Thank you, ChatGPT:
	std::string time_stamp_string() const
	{
		return std::format("{:%F %T}", std::chrono::floor<std::chrono::seconds>(time_stamp_));
	}

	int trade_id() const
	{
		return trade_id_;
	}


private:
	Buy_or_Sell buy_or_sell_;
	double price_;
	int num_shares_;
	std::chrono::time_point<std::chrono::system_clock> time_stamp_;
	inline static const double epsilon_ = std::sqrt(std::numeric_limits<double>::epsilon());
	inline static int trade_id_ = 0;		// Will be incremented
};
